"""Helper module for importing backported packages. So that the rest of the
modules is not cluttered with conditional imports.
"""

import sys

# We use version check instead of try/except because mypy does not like it.
if sys.version_info >= (3, 11):
    import tomllib  # noqa: F401
else:
    import tomli as tomllib  # noqa: F401



import os
from typing import Optional
from typing import Union
import urllib.parse

_WELL_KNOWN_PATH = ".ai.h2o.cloud.discovery"
_DEFAULT_LOCAL_CONFIG_PATH = "~/.h2oai/h2o-cli-config.toml"


class DetermineURIError(LookupError):
    """Raised when the discovery endpoint cannot be determined."""


def determine_uri(
    environment: Optional[str] = None,
    discovery_address: Optional[str] = None,
    config_endpoint: Optional[str] = None,
) -> str:
    """Uses passed parameters and environment variables to get the uri of the discovery
    service.
    """
    if discovery_address is not None and environment is not None:
        raise ValueError("cannot specify both discovery and environment")

    if discovery_address is not None:
        return discovery_address.rstrip("/")

    if environment is not None:
        return _discovery_uri_from_environment(environment)

    discovery_address = os.environ.get("H2O_CLOUD_DISCOVERY")
    if discovery_address is not None:
        return discovery_address.rstrip("/")

    environment = os.environ.get("H2O_CLOUD_ENVIRONMENT")
    if environment is not None:
        return _discovery_uri_from_environment(environment)

    if config_endpoint is not None:
        return _discovery_uri_from_environment(config_endpoint)

    raise DetermineURIError


def _discovery_uri_from_environment(environment: str):
    return urllib.parse.urljoin(environment + "/", _WELL_KNOWN_PATH)


def determine_local_config_path(
    config_path: Optional[Optional[Union[str, bytes, os.PathLike]]] = None,
) -> Optional[str]:
    """Uses passed parameter, environment variable and H2O CLI default to get the
    path to the local config file.
    """
    if config_path is not None:
        return str(os.fspath(config_path))

    config_path = os.environ.get("H2OCONFIG")
    if config_path is not None:
        return os.fspath(config_path)

    local_config_path = os.path.expanduser(_DEFAULT_LOCAL_CONFIG_PATH)
    if not os.path.isfile(local_config_path):
        return None
    return local_config_path



class DiscoveryError(Exception):
    """Base class for all exceptions of the H2O Cloud discovery Client."""


class DiscoveryLookupError(DiscoveryError, LookupError):
    """Raised when the discovery URI cannot be determined."""


class H2OCloudEnvironmentError(DiscoveryError):
    """Raised when there seems to be a problem with the deployment of the server
    side.
    """



import pytest

from h2o_discovery._internal import lookup


def environment_test_cases():
    yield pytest.param(
        "https://test.h2o.ai", "https://test.h2o.ai/.ai.h2o.cloud.discovery", id="clean"
    )
    yield pytest.param(
        "https://test.h2o.ai/",
        "https://test.h2o.ai/.ai.h2o.cloud.discovery",
        id="with trailing slash",
    )
    yield pytest.param(
        "https://test.h2o.ai:1234",
        "https://test.h2o.ai:1234/.ai.h2o.cloud.discovery",
        id="with port",
    )
    yield pytest.param(
        "https://test.h2o.ai:1234/",
        "https://test.h2o.ai:1234/.ai.h2o.cloud.discovery",
        id="with port and trailing slash",
    )
    yield pytest.param(
        "https://test.h2o.ai/path",
        "https://test.h2o.ai/path/.ai.h2o.cloud.discovery",
        id="with path",
    )
    yield pytest.param(
        "https://test.h2o.ai/path/",
        "https://test.h2o.ai/path/.ai.h2o.cloud.discovery",
        id="with path and trailing slash",
    )
    yield pytest.param(
        "https://test.h2o.ai:1234/path",
        "https://test.h2o.ai:1234/path/.ai.h2o.cloud.discovery",
        id="with path and port",
    )
    yield pytest.param(
        "https://test.h2o.ai:1234/path/",
        "https://test.h2o.ai:1234/path/.ai.h2o.cloud.discovery",
        id="with path, port and trailing slash",
    )


def discovery_test_cases():
    yield pytest.param(
        "http://test-service.domain:1234", "http://test-service.domain:1234", id="clean"
    )
    yield pytest.param(
        "http://test-service.domain:1234/",
        "http://test-service.domain:1234",
        id="with trailing slash",
    )
    yield pytest.param(
        "http://test-service.domain:1234/path",
        "http://test-service.domain:1234/path",
        id="with path",
    )
    yield pytest.param(
        "http://test-service.domain:1234/path/",
        "http://test-service.domain:1234/path",
        id="with path and trailing slash",
    )


@pytest.mark.parametrize("environment_input,expected_uri", environment_test_cases())
def test_determine_uri_environment_param(environment_input, expected_uri):
    # When
    uri = lookup.determine_uri(environment=environment_input)

    # Then
    assert uri == expected_uri


@pytest.mark.parametrize("environment_input,expected_uri", environment_test_cases())
def test_determine_uri_environment_env_var(
    monkeypatch, environment_input, expected_uri
):
    # Given
    monkeypatch.setenv("H2O_CLOUD_ENVIRONMENT", environment_input)

    # When
    uri = lookup.determine_uri()

    # Then
    assert uri == expected_uri


@pytest.mark.parametrize("discovery_input,expected_uri", discovery_test_cases())
def test_determine_uri_discovery_param(discovery_input, expected_uri):
    # When
    uri = lookup.determine_uri(discovery_address=discovery_input)

    # Then
    assert uri == expected_uri


@pytest.mark.parametrize("discovery_input,expected_uri", discovery_test_cases())
def test_determine_uri_discovery_env_var(monkeypatch, discovery_input, expected_uri):
    # Given
    monkeypatch.setenv("H2O_CLOUD_DISCOVERY", discovery_input)

    # When
    uri = lookup.determine_uri()

    # Then
    assert uri == expected_uri


def test_determine_uri_both_env_var_discovery_takes_precedence(monkeypatch):
    # Given
    monkeypatch.setenv("H2O_CLOUD_ENVIRONMENT", "https://test.h2o.ai")
    monkeypatch.setenv("H2O_CLOUD_DISCOVERY", "http://test-service.domain:1234")

    # When
    uri = lookup.determine_uri()

    # Then
    assert uri == "http://test-service.domain:1234"


def test_determine_uri_environment_param_takes_precedence(monkeypatch):
    # Given
    monkeypatch.setenv("H2O_CLOUD_ENVIRONMENT", "https://test-env.h2o.ai")
    environment = "https://test-param.h2o.ai"

    # When
    uri = lookup.determine_uri(environment=environment)

    # Then
    assert uri == "https://test-param.h2o.ai/.ai.h2o.cloud.discovery"


def test_determine_uri_discovery_param_takes_precedence(monkeypatch):
    # Given
    monkeypatch.setenv("H2O_CLOUD_DISCOVERY", "http://test-env.domain:1234")
    discovery = "http://test-param.domain:1234"

    # When
    uri = lookup.determine_uri(discovery_address=discovery)

    # Then
    assert uri == "http://test-param.domain:1234"


def test_determine_uri_environment_param_takes_precedence_over_discovery_env_var(
    monkeypatch,
):
    # Given
    monkeypatch.setenv("H2O_CLOUD_DISCOVERY", "http://test-env.domain:1234")
    environment = "https://test-param.h2o.ai"

    # When
    uri = lookup.determine_uri(environment=environment)

    # Then
    assert uri == "https://test-param.h2o.ai/.ai.h2o.cloud.discovery"


def test_determine_uri_cannot_set_both_params():
    # Given
    environment = "https://test.h2o.ai"
    discovery = "http://test-service.domain:1234"

    # When
    with pytest.raises(ValueError) as excinfo:
        lookup.determine_uri(environment=environment, discovery_address=discovery)

    # Then
    assert "cannot specify both discovery and environment" in str(excinfo.value)


def test_determine_uri_cannot_determine_url():
    # Given
    environment = None
    discovery = None

    # When / Then
    with pytest.raises(lookup.DetermineURIError):
        lookup.determine_uri(environment=environment, discovery_address=discovery)



import httpx
import pytest
import respx

from h2o_discovery import model
from h2o_discovery._internal import client


@respx.mock
def test_client_get_environment_internal():
    # Given
    route = respx.get("http://test.example.com:1234/v1/environment").respond(
        json=ENVIRONMENT_JSON
    )
    cl = client.Client("http://test.example.com:1234")

    # When
    env = cl.get_environment()

    # Then
    assert route.called
    assert env == EXPECTED_ENVIRONMENT_DATA


@respx.mock
def test_client_get_environment_public():
    # Given
    route = respx.get(
        "https://test.example.com/.ai.h2o.cloud.discovery/v1/environment"
    ).respond(json=ENVIRONMENT_JSON)
    cl = client.Client("https://test.example.com/.ai.h2o.cloud.discovery")

    # When
    env = cl.get_environment()

    # Then
    assert route.called
    assert env == EXPECTED_ENVIRONMENT_DATA


@respx.mock
@pytest.mark.asyncio
async def test_async_client_get_environment_internal():
    # Given
    route = respx.get("http://test.example.com:1234/v1/environment").respond(
        json=ENVIRONMENT_JSON
    )
    cl = client.AsyncClient("http://test.example.com:1234")

    # When
    env = await cl.get_environment()

    # Then
    assert route.called
    assert env == EXPECTED_ENVIRONMENT_DATA


@respx.mock
@pytest.mark.asyncio
async def test_async_client_get_environment_public():
    # Given
    route = respx.get(
        "https://test.example.com/.ai.h2o.cloud.discovery/v1/environment"
    ).respond(json=ENVIRONMENT_JSON)
    cl = client.AsyncClient("https://test.example.com/.ai.h2o.cloud.discovery")

    # When
    env = await cl.get_environment()

    # Then
    assert route.called
    assert env == EXPECTED_ENVIRONMENT_DATA


@respx.mock
def test_client_list_services_internal():
    # Given
    route = respx.get("http://test.example.com:1234/v1/services")
    route.side_effect = SERVICES_RESPONSES

    cl = client.Client("http://test.example.com:1234")

    # When
    services = cl.list_services()

    # Then
    assert services == EXPECTED_SERVICES_RECORDS
    _assert_pagination_api_calls(route)


@respx.mock
def test_client_list_services_public():
    # Given
    route = respx.get("https://test.example.com/.ai.h2o.cloud.discovery/v1/services")
    route.side_effect = SERVICES_RESPONSES

    cl = client.Client("https://test.example.com/.ai.h2o.cloud.discovery")

    # When
    services = cl.list_services()

    # Then
    assert services == EXPECTED_SERVICES_RECORDS
    _assert_pagination_api_calls(route)


@respx.mock
@pytest.mark.asyncio
async def test_async_client_list_services_internal():
    # Given
    route = respx.get("http://test.example.com:1234/v1/services")
    route.side_effect = SERVICES_RESPONSES

    cl = client.AsyncClient("http://test.example.com:1234")

    # When
    services = await cl.list_services()

    # Then
    assert services == EXPECTED_SERVICES_RECORDS
    _assert_pagination_api_calls(route)


@respx.mock
@pytest.mark.asyncio
async def test_async_client_list_services_public():
    # Given
    route = respx.get("https://test.example.com/.ai.h2o.cloud.discovery/v1/services")
    route.side_effect = SERVICES_RESPONSES

    cl = client.AsyncClient("https://test.example.com/.ai.h2o.cloud.discovery")

    # When
    services = await cl.list_services()

    # Then
    assert services == EXPECTED_SERVICES_RECORDS
    _assert_pagination_api_calls(route)


@respx.mock
def test_client_list_clients_internal():
    # Given
    route = respx.get("http://test.example.com:1234/v1/clients")
    route.side_effect = CLIENTS_RESPONSES

    cl = client.Client("http://test.example.com:1234")

    # When
    clients = cl.list_clients()

    # Then

    assert clients == EXPECTED_CLIENTS_RECORDS
    _assert_pagination_api_calls(route)


@respx.mock
def test_client_list_clients_public():
    # Given
    route = respx.get("https://test.example.com/.ai.h2o.cloud.discovery/v1/clients")
    route.side_effect = CLIENTS_RESPONSES

    cl = client.Client("https://test.example.com/.ai.h2o.cloud.discovery")

    # When
    clients = cl.list_clients()

    # Then

    assert clients == EXPECTED_CLIENTS_RECORDS
    _assert_pagination_api_calls(route)


@respx.mock
@pytest.mark.asyncio
async def test_async_client_list_clients_internal():
    # Given
    route = respx.get("https://test.example.com:1234/v1/clients")
    route.side_effect = CLIENTS_RESPONSES

    cl = client.AsyncClient("https://test.example.com:1234")

    # When
    clients = await cl.list_clients()

    # Then

    assert clients == EXPECTED_CLIENTS_RECORDS
    _assert_pagination_api_calls(route)


@respx.mock
@pytest.mark.asyncio
async def test_async_client_list_clients_public():
    # Given
    route = respx.get("https://test.example.com/.ai.h2o.cloud.discovery/v1/clients")
    route.side_effect = CLIENTS_RESPONSES

    cl = client.AsyncClient("https://test.example.com/.ai.h2o.cloud.discovery")

    # When
    clients = await cl.list_clients()

    # Then

    assert clients == EXPECTED_CLIENTS_RECORDS
    _assert_pagination_api_calls(route)


@respx.mock
def test_client_list_links_internal():
    # Given
    route = respx.get("http://test.example.com:1234/v1/links")
    route.side_effect = LINKS_RESPONSES

    cl = client.Client("http://test.example.com:1234")

    # When
    links = cl.list_links()

    # Then
    assert links == EXPECTED_LINKS_RECORDS
    _assert_pagination_api_calls(route)


@respx.mock
def test_client_list_links_public():
    # Given
    route = respx.get("https://test.example.com/.ai.h2o.cloud.discovery/v1/links")
    route.side_effect = LINKS_RESPONSES

    cl = client.Client("https://test.example.com/.ai.h2o.cloud.discovery")

    # When
    links = cl.list_links()

    # Then
    assert links == EXPECTED_LINKS_RECORDS
    _assert_pagination_api_calls(route)


@respx.mock
@pytest.mark.asyncio
async def test_async_client_list_links_internal():
    # Given
    route = respx.get("https://test.example.com:1234/v1/links")
    route.side_effect = LINKS_RESPONSES

    cl = client.AsyncClient("https://test.example.com:1234")

    # When
    links = await cl.list_links()

    # Then
    assert links == EXPECTED_LINKS_RECORDS
    _assert_pagination_api_calls(route)


@respx.mock
@pytest.mark.asyncio
async def test_async_client_list_links_public():
    # Given
    route = respx.get("https://test.example.com/.ai.h2o.cloud.discovery/v1/links")
    route.side_effect = LINKS_RESPONSES

    cl = client.AsyncClient("https://test.example.com/.ai.h2o.cloud.discovery")

    # When
    links = await cl.list_links()

    # Then
    assert links == EXPECTED_LINKS_RECORDS
    _assert_pagination_api_calls(route)


@respx.mock
def test_client_list_services_can_handle_empty_response():
    # Given
    respx.get("https://test.example.com/v1/services").respond(json={})
    cl = client.Client("https://test.example.com")

    # When
    services = cl.list_services()

    # Then
    assert services == []


@respx.mock
@pytest.mark.asyncio
async def test_async_client_list_services_can_handle_empty_response():
    # Given
    respx.get("https://test.example.com/v1/services").respond(json={})
    cl = client.AsyncClient("https://test.example.com")

    # When
    services = await cl.list_services()

    # Then
    assert services == []


@respx.mock
def test_client_list_clients_can_handle_empty_response():
    # Given
    respx.get("https://test.example.com/v1/clients").respond(json={})
    cl = client.Client("https://test.example.com")

    # When
    services = cl.list_clients()

    # Then
    assert services == []


@respx.mock
@pytest.mark.asyncio
async def test_async_client_list_clients_can_handle_empty_response():
    # Given
    respx.get("https://test.example.com/v1/clients").respond(json={})
    cl = client.AsyncClient("https://test.example.com")

    # When
    services = await cl.list_clients()

    # Then
    assert services == []


@respx.mock
def test_client_list_links_can_handle_empty_response():
    # Given
    respx.get("https://test.example.com/v1/links").respond(json={})
    cl = client.Client("https://test.example.com")

    # When
    links = cl.list_links()

    # Then
    assert links == []


@respx.mock
@pytest.mark.asyncio
async def test_async_client_list_links_can_handle_empty_response():
    # Given
    respx.get("https://test.example.com/v1/links").respond(json={})
    cl = client.AsyncClient("https://test.example.com")

    # When
    links = await cl.list_links()

    # Then
    assert links == []


def _assert_pagination_api_calls(route):
    assert route.call_count == 3
    assert not route.calls[0].request.url.query
    assert route.calls[1].request.url.query == b"pageToken=next-page-token-1"
    assert route.calls[2].request.url.query == b"pageToken=next-page-token-2"


ENVIRONMENT_JSON = {
    "environment": {
        "h2oCloudEnvironment": "https://cloud.fbi.com",
        "h2oCloudPlatformOauth2Scope": "openid profile email",
        "issuerUrl": "https://phantauth.net",
        "h2oCloudVersion": "70.25",
    }
}

EXPECTED_ENVIRONMENT_DATA = model.Environment(
    h2o_cloud_environment="https://cloud.fbi.com",
    h2o_cloud_platform_oauth2_scope="openid profile email",
    issuer_url="https://phantauth.net",
    h2o_cloud_version="70.25",
)

SERVICES_RESPONSES = [
    httpx.Response(
        200,
        json={
            "services": [
                {
                    "name": "services/test-service-1",
                    "displayName": "Test Service 1 ",
                    "uri": "http://test-service-1.domain:1234",
                    "version": "1.0.0",
                    "oauth2Scope": "test-service-scope-1",
                    "pythonClient": "test-client-1==1.0.0",
                },
                {
                    "name": "services/test-service-2",
                    "displayName": "Test Service 2",
                    "uri": "http://test-service-2.domain:1234",
                },
            ],
            "nextPageToken": "next-page-token-1",
        },
    ),
    httpx.Response(
        200,
        json={
            "services": [
                {
                    "name": "services/test-service-3",
                    "displayName": "Test Service 3",
                    "uri": "http://test-service-3.domain:1234",
                }
            ],
            "nextPageToken": "next-page-token-2",
        },
    ),
    httpx.Response(
        200,
        json={
            "services": [
                {
                    "name": "services/test-service-4",
                    "displayName": "Test Service 4",
                    "uri": "http://test-service-4.domain:1234",
                }
            ]
        },
    ),
]

EXPECTED_SERVICES_RECORDS = [
    model.Service(
        name="services/test-service-1",
        display_name="Test Service 1 ",
        uri="http://test-service-1.domain:1234",
        version="1.0.0",
        oauth2_scope="test-service-scope-1",
        python_client="test-client-1==1.0.0",
    ),
    model.Service(
        name="services/test-service-2",
        display_name="Test Service 2",
        uri="http://test-service-2.domain:1234",
    ),
    model.Service(
        name="services/test-service-3",
        display_name="Test Service 3",
        uri="http://test-service-3.domain:1234",
    ),
    model.Service(
        name="services/test-service-4",
        display_name="Test Service 4",
        uri="http://test-service-4.domain:1234",
    ),
]

CLIENTS_RESPONSES = [
    httpx.Response(
        200,
        json={
            "clients": [
                {
                    "name": "clients/test-client-1",
                    "displayName": "Test Client 1",
                    "oauth2ClientId": "test-client-1",
                },
                {
                    "name": "clients/test-client-2",
                    "displayName": "Test Client 2",
                    "oauth2ClientId": "test-client-2",
                },
            ],
            "nextPageToken": "next-page-token-1",
        },
    ),
    httpx.Response(
        200,
        json={
            "clients": [
                {
                    "name": "clients/test-client-3",
                    "displayName": "Test Client 3",
                    "oauth2ClientId": "test-client-3",
                }
            ],
            "nextPageToken": "next-page-token-2",
        },
    ),
    httpx.Response(
        200,
        json={
            "clients": [
                {
                    "name": "clients/test-client-4",
                    "displayName": "Test Client 4",
                    "oauth2ClientId": "test-client-4",
                }
            ]
        },
    ),
]

EXPECTED_CLIENTS_RECORDS = [
    model.Client(
        name="clients/test-client-1",
        display_name="Test Client 1",
        oauth2_client_id="test-client-1",
    ),
    model.Client(
        name="clients/test-client-2",
        display_name="Test Client 2",
        oauth2_client_id="test-client-2",
    ),
    model.Client(
        name="clients/test-client-3",
        display_name="Test Client 3",
        oauth2_client_id="test-client-3",
    ),
    model.Client(
        name="clients/test-client-4",
        display_name="Test Client 4",
        oauth2_client_id="test-client-4",
    ),
]


LINKS_RESPONSES = [
    httpx.Response(
        200,
        json={
            "links": [
                {"name": "links/test-link-1", "uri": "http://test-link-1.domain:1234"},
                {
                    "name": "links/test-link-2",
                    "uri": "http://test-link-2.domain:1234",
                    "text": "Test Link 2",
                },
            ],
            "nextPageToken": "next-page-token-1",
        },
    ),
    httpx.Response(
        200,
        json={
            "links": [
                {
                    "name": "links/test-link-3",
                    "uri": "http://test-link-3.domain:1234",
                    "text": "Test Link 3",
                }
            ],
            "nextPageToken": "next-page-token-2",
        },
    ),
    httpx.Response(
        200,
        json={
            "links": [
                {"name": "links/test-link-4", "uri": "http://test-link-4.domain:1234"}
            ]
        },
    ),
]

EXPECTED_LINKS_RECORDS = [
    model.Link(
        name="links/test-link-1", uri="http://test-link-1.domain:1234", text=None
    ),
    model.Link(
        name="links/test-link-2",
        uri="http://test-link-2.domain:1234",
        text="Test Link 2",
    ),
    model.Link(
        name="links/test-link-3",
        uri="http://test-link-3.domain:1234",
        text="Test Link 3",
    ),
    model.Link(
        name="links/test-link-4", uri="http://test-link-4.domain:1234", text=None
    ),
]



import dataclasses
import datetime
import os
import ssl
from typing import Optional
from typing import Tuple
from typing import Union

from h2o_discovery import error
from h2o_discovery import model
from h2o_discovery._internal import client
from h2o_discovery._internal import config
from h2o_discovery._internal import load
from h2o_discovery._internal import lookup
from h2o_discovery._version import __version__  # noqa: F401

Discovery = model.Discovery


def discover(
    environment: Optional[str] = None,
    discovery_address: Optional[str] = None,
    config_path: Optional[Union[str, bytes, os.PathLike]] = None,
    *,
    http_timeout: Optional[datetime.timedelta] = None,
    ssl_context: Optional[ssl.SSLContext] = None,
) -> Discovery:
    """Obtains and returns a Discovery object from the discovery service.

    All arguments are optional. Discovery determined with the following precedence:
        - discovery_address parameter
        - H2O_CLOUD_DISCOVERY environment variable
        - environment parameter
        - H2O_CLOUD_ENVIRONMENT environment variable
        - environment URI loaded from the local H2O CLI configuration file

    Config path is determined with the following precedence:
        - config_path parameter
        - H2OCONFIG environment variable
        - default H2O CLI configuration configuration path
        - if default path does not exist, no config is loaded

    Args:
        environment: The H2O Cloud environment URL to use (e.g. https://cloud.h2o.ai).
        discovery_address: The address of the discovery service.
        config_path: The path to the H2O CLI configuration file.
        http_timeout: The timeout for HTTP requests. Value applies to all of the
            timeouts (connect, read, write).
        ssl_context: The SSL context to use for HTTPS requests.
            If not specified default SSL context is used.


    Raises:
        exceptions.DiscoveryLookupError: If the URI cannot be determined.
        exceptions.H2OCloudEnvironmentError: If there seems to be a problem with the
            deployment of the server side.
    """
    uri, cfg = _lookup_and_load(environment, discovery_address, config_path)

    discovery = load.load_discovery(
        client.Client(uri=uri, timeout=http_timeout, ssl_context=ssl_context)
    )
    credentials = load.load_credentials(
        clients=discovery.clients, config_tokens=cfg.tokens
    )

    return dataclasses.replace(discovery, credentials=credentials)


async def discover_async(
    environment: Optional[str] = None,
    discovery_address: Optional[str] = None,
    config_path: Optional[Union[str, bytes, os.PathLike]] = None,
    *,
    http_timeout: Optional[datetime.timedelta] = None,
    ssl_context: Optional[ssl.SSLContext] = None,
) -> Discovery:
    """Obtains and returns a Discovery object from the discovery service.

    All arguments are optional. Discovery determined with the following precedence:
        - discovery_address parameter
        - H2O_CLOUD_DISCOVERY environment variable
        - environment parameter
        - H2O_CLOUD_ENVIRONMENT environment variable
        - environment URI loaded from the local H2O CLI configuration file

    Config path is determined with the following precedence:
        - config_path parameter
        - H2OCONFIG environment variable
        - default H2O CLI configuration configuration path
        - if default path does not exist, no config is loaded

    Args:
        environment: The H2O Cloud environment URL to use (e.g. https://cloud.h2o.ai).
        discovery_address: The address of the discovery service.
        config_path: The path to the H2O CLI configuration file.
        http_timeout: The timeout for HTTP requests. Value applies to all of the
            timeouts (connect, read, write).
        ssl_context: The SSL context to use for HTTPS requests.
            If not specified default SSL context is used.

    Raises:
        exceptions.DiscoveryLookupError: If the URI cannot be determined.
        exceptions.H2OCloudEnvironmentError: If there seems to be a problem with the
            deployment of the server side.
    """

    uri, cfg = _lookup_and_load(environment, discovery_address, config_path)

    discovery = await load.load_discovery_async(
        client.AsyncClient(uri=uri, timeout=http_timeout, ssl_context=ssl_context)
    )
    credentials = load.load_credentials(
        clients=discovery.clients, config_tokens=cfg.tokens
    )

    return dataclasses.replace(discovery, credentials=credentials)


def _lookup_and_load(
    environment: Optional[str] = None,
    discovery_address: Optional[str] = None,
    config_path: Optional[Union[str, bytes, os.PathLike]] = None,
) -> Tuple[str, config.Config]:
    cfg = config.load_config(lookup.determine_local_config_path(config_path))

    try:
        uri = lookup.determine_uri(environment, discovery_address, cfg.endpoint)
    except lookup.DetermineURIError:
        raise error.DiscoveryLookupError(
            "Cannot determine discovery URI."
            " Please set H2O_CLOUD_ENVIRONMENT or H2O_CLOUD_DISCOVERY environment"
            " variables or use the environment or discovery parameters."
            " Alternatively, you can create configure your local environment with"
            " the with H2O CLI with and/or use H2OCONFIG environment variable"
            " (see https://docs.h2o.ai/h2o-ai-cloud/developerguide/cli)."
        ) from None

    return uri, cfg



from h2o_discovery._internal import lookup


def test_determine_local_config_path_param():
    # When
    path = lookup.determine_local_config_path("test-path")

    # Then
    assert path == "test-path"


def test_determine_local_config_path_env(monkeypatch):
    # Given
    monkeypatch.setenv("H2OCONFIG", "test-path-from-env")

    # When
    path = lookup.determine_local_config_path()

    # Then
    assert path == "test-path-from-env"


def test_determine_local_config_path_default_location_does_not_exist(
    monkeypatch, tmp_path
):
    # Given
    home = tmp_path / "home" / "test-user"
    home.mkdir(parents=True)
    monkeypatch.setenv("HOME", str(home))

    # When
    path = lookup.determine_local_config_path()

    # Then
    assert path is None


def test_determine_local_config_path_default_location_does_exists(
    monkeypatch, tmp_path
):
    # Given
    home = tmp_path / "home" / "test-user"
    home.mkdir(parents=True)
    monkeypatch.setenv("HOME", str(home))

    config_dir = home / ".h2oai"
    config_dir.mkdir(parents=True)
    config_file = config_dir / "h2o-cli-config.toml"
    config_file.touch()

    # When
    path = lookup.determine_local_config_path()

    # Then
    assert path == str(config_file)



import pytest

from h2o_discovery._internal import config


def config_test_cases():
    yield pytest.param("", config.Config(endpoint=None, tokens={}), id="empty config")
    yield pytest.param(
        """
        Endpoint = "https://cloud.h2o.ai"
        ClientID = "client-id"
        Token = "TestToken"
        PlatformClientID = "platform-client-id"
        PlatformToken = "TestPlatformToken"
        """,
        config.Config(
            endpoint="https://cloud.h2o.ai",
            tokens={
                "client-id": "TestToken",
                "platform-client-id": "TestPlatformToken",
            },
        ),
        id="full config",
    )
    yield pytest.param(
        """
        ClientID = "client-id"
        Token = "TestToken"
        """,
        config.Config(tokens={"client-id": "TestToken"}),
        id="cli token only",
    )
    yield pytest.param(
        """
        PlatformClientID = "platform-client-id"
        PlatformToken = "TestPlatformToken"
        """,
        config.Config(tokens={"platform-client-id": "TestPlatformToken"}),
        id="platform token only",
    )
    yield pytest.param(
        """
        Endpoint = "https://cloud.h2o.ai"
        """,
        config.Config(endpoint="https://cloud.h2o.ai", tokens={}),
        id="endpoint only",
    )


@pytest.mark.parametrize("config_content,expected_results", config_test_cases())
def test_load_config(config_content, expected_results, tmp_path):
    # Given
    config_file = tmp_path / "config.toml"
    config_file.write_text(config_content)

    # When
    result = config.load_config(config_file)

    # Then
    assert result == expected_results


def test_load_config_no_path(tmp_path):
    # When
    result = config.load_config()

    # Then
    assert result == config.Config()



import asyncio
import json
import os
import types
from typing import Iterable
from typing import List
from typing import Mapping
from typing import Optional

import httpx

from h2o_discovery import error
from h2o_discovery import model
from h2o_discovery._internal import client


def load_discovery(cl: client.Client) -> model.Discovery:
    """Loads the discovery records from the Discovery Service."""
    try:
        environment = cl.get_environment()
        services = _get_service_map(cl.list_services())
        clients = _get_client_map(cl.list_clients())
        links = _get_link_map(_list_links(cl))
        components = _get_component_map(_list_components(cl))
    except Exception as e:
        _handle_specific_client_exceptions(e)
        raise
    return model.Discovery(
        environment=environment,
        services=services,
        clients=clients,
        links=links,
        components=components,
    )


async def load_discovery_async(cl: client.AsyncClient) -> model.Discovery:
    """Loads the discovery records from the Discovery Service."""
    try:
        return await _gather_discovery_async(cl)
    except Exception as e:
        _handle_specific_client_exceptions(e)
        raise


async def _gather_discovery_async(cl: client.AsyncClient) -> model.Discovery:
    environment_future = cl.get_environment()
    services_future = cl.list_services()
    clients_future = cl.list_clients()
    links_future = _list_links_async(cl)
    components_future = _list_components_async(cl)

    environment, services, clients, links, components = await asyncio.gather(
        environment_future,
        services_future,
        clients_future,
        links_future,
        components_future,
    )

    return model.Discovery(
        environment=environment,
        services=_get_service_map(services),
        clients=_get_client_map(clients),
        links=_get_link_map(links),
        components=_get_component_map(components),
    )


def _list_links(cl: client.Client) -> List[model.Link]:
    try:
        return cl.list_links()
    except Exception as e:
        # Links are added in the server version 2.5.0. In order to have client that
        # is backwards compatible, we won't fail if the links are not available.
        if isinstance(e, httpx.HTTPStatusError) and e.response.status_code == 404:
            return []
        raise


async def _list_links_async(cl: client.AsyncClient) -> List[model.Link]:
    try:
        return await cl.list_links()
    except Exception as e:
        # Links are added in the server version 2.5.0. In order to have client that
        # is backwards compatible, we won't fail if the links are not available.
        if isinstance(e, httpx.HTTPStatusError) and e.response.status_code == 404:
            return []
        raise


def _list_components(cl: client.Client) -> List[model.Component]:
    try:
        return cl.list_components()
    except Exception as e:
        # Components are added in the server version 2.6.0. In order to have client that
        # is backwards compatible, we won't fail if the components are not available.
        if isinstance(e, httpx.HTTPStatusError) and e.response.status_code == 404:
            return []
        raise


async def _list_components_async(cl: client.AsyncClient) -> List[model.Component]:
    try:
        return await cl.list_components()
    except Exception as e:
        # Components are added in the server version 2.6.0. In order to have client that
        # is backwards compatible, we won't fail if the components are not available.
        if isinstance(e, httpx.HTTPStatusError) and e.response.status_code == 404:
            return []
        raise


def load_credentials(
    clients: Mapping[str, model.Client], config_tokens: Optional[Mapping[str, str]]
) -> Mapping[str, model.Credentials]:
    """Loads client credentials from the environment or tokens loaded from the
    config.
    """
    tokens: Mapping[str, str] = {}
    if config_tokens is not None:
        tokens = config_tokens

    out = {}
    for name, cl in clients.items():
        env_name = f"H2O_CLOUD_CLIENT_{name.upper()}_TOKEN"
        token = os.environ.get(env_name) or tokens.get(cl.oauth2_client_id)
        if token:
            out[name] = model.Credentials(client=name, refresh_token=token)
    return types.MappingProxyType(out)


def _get_service_map(services: Iterable[model.Service]) -> Mapping[str, model.Service]:
    out = {}
    for s in services:
        out[_service_key(s.name)] = s
    return types.MappingProxyType(out)


def _get_client_map(clients: Iterable[model.Client]) -> Mapping[str, model.Client]:
    out = {}
    for c in clients:
        out[_client_key(c.name)] = c
    return types.MappingProxyType(out)


def _get_link_map(links: Iterable[model.Link]) -> Mapping[str, model.Link]:
    out = {}
    for ln in links:
        out[_link_key(ln.name)] = ln
    return types.MappingProxyType(out)


def _get_component_map(
    components: Iterable[model.Component],
) -> Mapping[str, model.Component]:
    out = {}
    for c in components:
        out[_component_key(c.name)] = c
    return types.MappingProxyType(out)


_SERVICES_COLLECTION_PREFIX = "services/"


def _service_key(name: str) -> str:
    if name.startswith(_SERVICES_COLLECTION_PREFIX):
        return name[len(_SERVICES_COLLECTION_PREFIX) :]
    raise ValueError(f"invalid service name: {name}")


_CLIENTS_COLLECTION_PREFIX = "clients/"


def _client_key(name: str) -> str:
    if name.startswith(_CLIENTS_COLLECTION_PREFIX):
        return name[len(_CLIENTS_COLLECTION_PREFIX) :]
    raise ValueError(f"invalid client name: {name}")


_LINKS_COLLECTION_PREFIX = "links/"


def _link_key(name: str) -> str:
    if name.startswith(_LINKS_COLLECTION_PREFIX):
        return name[len(_LINKS_COLLECTION_PREFIX) :]
    raise ValueError(f"invalid link name: {name}")


_COMPONENTS_COLLECTION_PREFIX = "components/"


def _component_key(name: str) -> str:
    if name.startswith(_COMPONENTS_COLLECTION_PREFIX):
        return name[len(_COMPONENTS_COLLECTION_PREFIX) :]
    raise ValueError(f"invalid component name: {name}")


_ENV_ERROR = error.H2OCloudEnvironmentError(
    "Received an unexpected response from the server."
    " Please make sure that the environment you are trying to connect to is"
    " configured correctly and has the H2O Cloud Discovery Service enabled."
)


def _handle_specific_client_exceptions(e: Exception) -> None:
    if isinstance(e, json.decoder.JSONDecodeError):
        raise _ENV_ERROR from e
    if isinstance(e, httpx.HTTPStatusError) and e.response.status_code == 404:
        raise _ENV_ERROR from e



from h2o_discovery import model


def test_service_from_json_dict():
    # Given
    json = {
        "name": "services/test-service",
        "displayName": "Test Service",
        "uri": "http://test-service.domain:1234",
        "version": "1.0.0",
        "oauth2Scope": "test-service-scope",
        "pythonClient": "test-client==1.0.0",
    }

    # When
    result = model.Service.from_json_dict(json)

    # Then
    assert result == model.Service(
        name="services/test-service",
        display_name="Test Service",
        uri="http://test-service.domain:1234",
        version="1.0.0",
        oauth2_scope="test-service-scope",
        python_client="test-client==1.0.0",
    )


def test_client_from_json_dict():
    # Given
    json = {
        "name": "test-client",
        "displayName": "Test Client",
        "oauth2ClientId": "test-client-id",
    }

    # When
    result = model.Client.from_json_dict(json)

    # Then
    assert result == model.Client(
        name="test-client",
        display_name="Test Client",
        oauth2_client_id="test-client-id",
    )


def test_environment_from_json_dict():
    # Given
    json = {
        "h2oCloudEnvironment": "https://test.h2o.ai",
        "issuerUrl": "https://test.h2o.ai",
        "h2oCloudPlatformOauth2Scope": "test-platform-scope",
        "h2oCloudVersion": "test-version",
    }

    # When
    result = model.Environment.from_json_dict(json)

    # Then
    assert result == model.Environment(
        h2o_cloud_environment="https://test.h2o.ai",
        issuer_url="https://test.h2o.ai",
        h2o_cloud_platform_oauth2_scope="test-platform-scope",
        h2o_cloud_version="test-version",
    )


def test_environment_from_json_dict_with_missing_version():
    # Given
    json = {
        "h2oCloudEnvironment": "https://test.h2o.ai",
        "issuerUrl": "https://test.h2o.ai",
        "h2oCloudPlatformOauth2Scope": "test-platform-scope",
    }

    # When
    result = model.Environment.from_json_dict(json)

    # Then
    assert result == model.Environment(
        h2o_cloud_environment="https://test.h2o.ai",
        issuer_url="https://test.h2o.ai",
        h2o_cloud_platform_oauth2_scope="test-platform-scope",
    )


def test_link_from_json_dictt():
    # Given
    json = {
        "name": "links/test-link",
        "uri": "http://test-link.domain:1234",
        "text": "Test Link",
    }

    # When
    result = model.Link.from_json_dict(json)

    # Then
    assert result == model.Link(
        name="links/test-link", uri="http://test-link.domain:1234", text="Test Link"
    )


def test_link_from_json_dict_with_missing_text():
    # Given
    json = {"name": "links/test-link", "uri": "http://test-link.domain:1234"}

    # When
    result = model.Link.from_json_dict(json)

    # Then
    assert result == model.Link(
        name="links/test-link", uri="http://test-link.domain:1234", text=None
    )


def test_discovery_without_new_entities_ok():
    """Test that Discovery can be created without explicitly providing links in
    the constructor to ensure backward compatibility.
    """

    # When
    _ = model.Discovery(
        environment=model.Environment(
            h2o_cloud_environment="https://test.h2o.ai",
            issuer_url="https://test.h2o.ai",
            h2o_cloud_platform_oauth2_scope="test-platform-scope",
        ),
        credentials={},
        services={},
        clients={},
        # No links nor components.
    )

    # Then
    # No exception is raised.



from h2o_discovery import model
from h2o_discovery._internal import load


def test_load_credentials(monkeypatch):
    # Given
    clients = {
        "env_client": model.Client(
            name="clients/env_client",
            display_name="Env Client",
            oauth2_client_id="env-client-id",
        ),
        "config_client": model.Client(
            name="clients/config_client",
            display_name="Config Client",
            oauth2_client_id="config-client-id",
        ),
        "extra_client": model.Client(
            name="clients/extra_client",
            display_name="Extra Client",
            oauth2_client_id="extra-client-id",
        ),
    }
    tokens = {"config-client-id": "config-token"}
    monkeypatch.setenv("H2O_CLOUD_CLIENT_ENV_CLIENT_TOKEN", "env-token")

    # When
    result = load.load_credentials(clients=clients, config_tokens=tokens)

    # Then
    assert result == {
        "env_client": model.Credentials(client="env_client", refresh_token="env-token"),
        "config_client": model.Credentials(
            client="config_client", refresh_token="config-token"
        ),
    }



import datetime
import ssl
from typing import Callable
from typing import List
from typing import Optional
from typing import TypeVar

import httpx

from h2o_discovery import model

_ENVIRONMENT_ENDPOINT = "v1/environment"
_SERVICES_ENDPOINT = "v1/services"
_CLIENTS_ENDPOINT = "v1/clients"
_LINKS_ENDPOINT = "v1/links"
_COMPONENTS_ENDPOINT = "v1/components"

DEFAULT_HTTP_TIMEOUT = datetime.timedelta(seconds=5)


class _BaseClient:
    def __init__(
        self,
        uri: str,
        timeout: Optional[datetime.timedelta] = None,
        ssl_context: Optional[ssl.SSLContext] = None,
    ):
        self._uri = uri
        self._verify = ssl_context or ssl.create_default_context()

        self._timeout = 5.0
        if timeout is not None:
            self._timeout = timeout.total_seconds()


_T = TypeVar("_T")


class Client(_BaseClient):
    """Synchronous Implementation of the Discovery Service API.

    Listing methods do pagination and always return all of the available objects.
    """

    def get_environment(self) -> model.Environment:
        """Returns the information about the environment."""
        with self._client() as client:
            resp = _fetch(client, _ENVIRONMENT_ENDPOINT)
            return model.Environment.from_json_dict(resp.json()["environment"])

    def list_services(self) -> List[model.Service]:
        """Returns the list of all registered services."""
        return self._get_all_entities(
            _SERVICES_ENDPOINT, "services", model.Service.from_json_dict
        )

    def list_clients(self) -> List[model.Client]:
        """Returns the list of all registered clients."""
        return self._get_all_entities(
            _CLIENTS_ENDPOINT, "clients", model.Client.from_json_dict
        )

    def list_links(self) -> List[model.Link]:
        """Returns the list of all registered links."""
        return self._get_all_entities(
            _LINKS_ENDPOINT, "links", model.Link.from_json_dict
        )

    def list_components(self) -> List[model.Component]:
        """Returns the list of all registered components."""
        return self._get_all_entities(
            _COMPONENTS_ENDPOINT, "components", model.Component.from_json_dict
        )

    def _get_all_entities(
        self, endpoint: str, collection_key: str, factory: Callable[[dict], _T]
    ) -> List[_T]:
        with self._client() as client:
            entities: List[_T] = []

            pages = _get_all_pages(client, endpoint)
            for page in pages:
                entities.extend([factory(d) for d in page.get(collection_key, [])])
            return entities

    def _client(self) -> httpx.Client:
        return httpx.Client(
            base_url=self._uri, timeout=self._timeout, verify=self._verify
        )


def _fetch(
    client: httpx.Client, uri: str, next_page_token: Optional[str] = None
) -> httpx.Response:
    params = None
    if next_page_token is not None:
        params = {"pageToken": next_page_token}
    resp = client.get(uri, params=params)
    resp.raise_for_status()
    return resp


def _get_all_pages(client: httpx.Client, uri: str) -> List[dict]:
    next_page_token: Optional[str] = None
    all_pages: List[dict] = []

    while True:
        resp = _fetch(client, uri, next_page_token)
        resp_json = resp.json()
        all_pages.append(resp_json)
        next_page_token = resp_json.get("nextPageToken")
        if next_page_token is None:
            return all_pages


class AsyncClient(_BaseClient):
    """Asynchronous Implementation of the Discovery Service API.

    Listing methods do pagination and always return all of the available objects.
    """

    async def get_environment(self) -> model.Environment:
        """Returns the information about the environment."""
        async with self._client() as client:
            resp = await _fetch_async(client, _ENVIRONMENT_ENDPOINT)
            return model.Environment.from_json_dict(resp.json()["environment"])

    async def list_services(self) -> List[model.Service]:
        """Returns the list of all registered services."""
        return await self._get_all_entities(
            _SERVICES_ENDPOINT, "services", model.Service.from_json_dict
        )

    async def list_clients(self) -> List[model.Client]:
        """Returns the list of all registered clients."""
        return await self._get_all_entities(
            _CLIENTS_ENDPOINT, "clients", model.Client.from_json_dict
        )

    async def list_links(self) -> List[model.Link]:
        """Returns the list of all registered links."""
        return await self._get_all_entities(
            _LINKS_ENDPOINT, "links", model.Link.from_json_dict
        )

    async def list_components(self) -> List[model.Component]:
        """Returns the list of all registered components."""
        return await self._get_all_entities(
            _COMPONENTS_ENDPOINT, "components", model.Component.from_json_dict
        )

    async def _get_all_entities(
        self, endpoint: str, collection_key: str, factory: Callable[[dict], _T]
    ) -> List[_T]:
        async with self._client() as client:
            entities: List[_T] = []

            pages = await _get_all_pages_async(client, endpoint)
            for page in pages:
                entities.extend([factory(d) for d in page.get(collection_key, [])])
            return entities

    def _client(self) -> httpx.AsyncClient:
        return httpx.AsyncClient(
            base_url=self._uri, timeout=self._timeout, verify=self._verify
        )


async def _fetch_async(
    client: httpx.AsyncClient, uri: str, next_page_token: Optional[str] = None
) -> httpx.Response:
    params = None
    if next_page_token is not None:
        params = {"pageToken": next_page_token}
    resp = await client.get(uri, params=params)
    resp.raise_for_status()
    return resp


async def _get_all_pages_async(client: httpx.AsyncClient, uri: str) -> List[dict]:
    next_page_token: Optional[str] = None
    all_pages: List[dict] = []

    while True:
        resp = await _fetch_async(client, uri, next_page_token)
        resp_json = resp.json()
        all_pages.append(resp_json)
        next_page_token = resp_json.get("nextPageToken")
        if next_page_token is None:
            return all_pages



import dataclasses
import types
from typing import Mapping
from typing import Optional


@dataclasses.dataclass(frozen=True)
class Service:
    """Representation of a registered service record."""

    #: Name of the Service. For example: "services/my-service-name".
    name: str

    #: Name of the Service that can be displayed on the front-end
    display_name: str

    #: URI for accessing the Service.
    #: This is usually the string that can be passed connection definition to the
    #: client for the particular service.
    uri: str

    #: Version of the service.
    #: Can be the version of the API or the version of
    #: the service. Clients can utilize this information change their behavior
    #: in accessing the service or downloading the correct client version.
    version: Optional[str] = None

    #: OAuth 2.0 Scope required to access the service.
    #: Clients request the access token with this scope in order to access the service.
    #: If the scope is not defined (or empty), clients should use
    #: h2o_cloud_platform_scope.
    oauth2_scope: Optional[str] = None

    #: Requirement Specifier (PEP 508) for the Python client that can be used for
    #: accessing the service.
    #: Any string that can be `pip install`ed.
    #:     Example: my-client==0.1.0
    python_client: Optional[str] = None

    @classmethod
    def from_json_dict(cls, json: Mapping[str, str]) -> "Service":
        """Create a Service from a JSON dict returned by the server."""
        return cls(
            name=json["name"],
            display_name=json["displayName"],
            uri=json["uri"],
            version=json.get("version"),
            oauth2_scope=json.get("oauth2Scope"),
            python_client=json.get("pythonClient"),
        )


@dataclasses.dataclass(frozen=True)
class Client:
    """Representation of a registered client record."""

    #: Name of the Client. For example: "clients/h2o-public-client".
    name: str

    #: Name of the Client that can be displayed on the front-end.
    display_name: str

    #: Public OAuth 2.0 client ID that the client needs to use to identify
    #: itself with the IDP.
    oauth2_client_id: str

    @classmethod
    def from_json_dict(cls, json: Mapping[str, str]) -> "Client":
        """Create a Client from a JSON dict returned by the server."""
        return cls(
            name=json["name"],
            display_name=json["displayName"],
            oauth2_client_id=json["oauth2ClientId"],
        )


@dataclasses.dataclass(frozen=True)
class Environment:
    """Representation of the information about the H2O Cloud environment."""

    #: Identifier of the environment. For example: "https://cloud.h2o.ai".
    #: This is the base URL of the environment. Clients can use this to validate
    #: that they are talking to the correct environment.
    h2o_cloud_environment: str

    #: OpenID Connect issuer_url.
    #: This is where clients find the OpenID Connect discovery on the well-known e
    #: endpoint.
    issuer_url: str

    #: OAuth 2.0 scope that clients should use to access the H2O Cloud Platform.
    #: This is the default scope that clients should use if the service does not
    #: define its own scope.
    h2o_cloud_platform_oauth2_scope: str

    #: Version of the H2O Cloud Platform release that is running in the environment.
    #: In the XC YY.MM.V format for released versions (e.g. MC 23.04.01 for managed
    #: cloud or HC 23.01.1 for hybrid cloud). Can be arbitrary string for
    #: testing or special environments.
    h2o_cloud_version: Optional[str] = None

    @classmethod
    def from_json_dict(cls, json: Mapping[str, str]) -> "Environment":
        """Create an Environment from a JSON dict returned by the server."""
        return cls(
            h2o_cloud_environment=json["h2oCloudEnvironment"],
            issuer_url=json["issuerUrl"],
            h2o_cloud_platform_oauth2_scope=json["h2oCloudPlatformOauth2Scope"],
            h2o_cloud_version=json.get("h2oCloudVersion"),
        )


@dataclasses.dataclass(frozen=True)
class Link:
    """Representation of the navigation link."""

    #: Canonical name of the Link. For example: "links/app-link".
    name: str

    #: Value that applications can put into the HTML anchor's href
    #: attribute to navigate to link.
    uri: str

    #: Preferred text to use as a link text. This usually make sense
    #: only as part of the navigation menus.
    #:     Example: <a href="{{ link.uri }}">{{ link.text }}</a>
    text: Optional[str]

    @classmethod
    def from_json_dict(cls, json: Mapping[str, str]) -> "Link":
        """Create a Link from a JSON dict returned by the server."""
        return cls(name=json["name"], uri=json["uri"], text=json.get("text"))


@dataclasses.dataclass(frozen=True)
class Component:
    """Representation of a registered component record."""

    #: Name of the Component. For example: "components/h2o-component".
    name: str

    #: Name of the Component that can be displayed on the front-end.
    display_name: str

    #: Version of the component.
    version: str

    #: Description of the Component. This will be displayed on the front-end.
    #: This should explain what the component does and why it belongs
    #: in the environment.
    description: Optional[str] = None

    @classmethod
    def from_json_dict(cls, json: Mapping[str, str]) -> "Component":
        """Create a Component from a JSON dict returned by the server."""
        return cls(
            name=json["name"],
            display_name=json["displayName"],
            version=json["version"],
            description=json.get("description"),
        )


@dataclasses.dataclass(frozen=True)
class Credentials:
    """Contain credentials associated with single registered client.

    Credentials are only determined locally and are not returned by the server.
    """

    #: Client identifier that the credentials are associated with.
    client: str

    #: Opaque string containing refresh token that can be used to obtain access token.
    refresh_token: str = dataclasses.field(
        # We don't want to show the refresh token to accidentally leak when printing
        # the object.
        repr=False
    )


def _empty_credentials_factory() -> Mapping[str, Credentials]:
    return types.MappingProxyType({})


def _empty_links_factory() -> Mapping[str, Link]:
    return types.MappingProxyType({})


def _empty_components_factory() -> Mapping[str, Component]:
    return types.MappingProxyType({})


@dataclasses.dataclass(frozen=True)
class Discovery:
    """Representation of the discovery records."""

    #: Information about the environment.
    environment: Environment

    #: Map of registered services in the `{"service-identifier": Service(...)}` format.
    services: Mapping[str, Service]

    #: Map of registered clients in the `{"client-identifier": Client(...)}` format.
    clients: Mapping[str, Client]

    #: Map of registered links in the `{"link-identifier": Link(...)}` format.
    links: Mapping[str, Link] = dataclasses.field(default_factory=_empty_links_factory)

    #: Map of registered components in the
    #: `{"component-identifier": Component(...)}` format.
    components: Mapping[str, Component] = dataclasses.field(
        default_factory=_empty_components_factory
    )

    #: Map of credentials in the `{"client-identifier": Credentials(...)}` format.
    credentials: Mapping[str, Credentials] = dataclasses.field(
        default_factory=_empty_credentials_factory
    )



import dataclasses
import types
from typing import Mapping
from typing import Optional

from h2o_discovery._internal.compat import tomllib


def _empty_tokens_factory() -> Mapping[str, str]:
    return types.MappingProxyType({})


@dataclasses.dataclass(frozen=True)
class Config:
    """Internal representation of the H2O CLI Configuration."""

    #: Configured URI of environment.
    endpoint: Optional[str] = None

    #: Map of found tokens in the configuration file. in the `{"client-id": "token"}`
    #: format.
    tokens: Mapping[str, str] = dataclasses.field(default_factory=_empty_tokens_factory)


def load_config(path: Optional[str] = None) -> Config:
    """Loads the H2O CLI config from the specified path.

    If no path is specified, an empty configuration is returned.
    """

    if not path:
        return Config()

    with open(path, "rb") as f:
        data = tomllib.load(f)

    endpoint = data.get("Endpoint")
    client_id = data.get("ClientID")
    token = data.get("Token")
    platform_client_id = data.get("PlatformClientID")
    platform_token = data.get("PlatformToken")

    tokens = {}
    if client_id is not None and token is not None:
        tokens[client_id] = token
    if platform_client_id is not None and platform_token is not None:
        tokens[platform_client_id] = platform_token

    return Config(endpoint=endpoint, tokens=types.MappingProxyType(tokens))



import asyncio
from unittest import mock

import httpx
import pytest

from h2o_discovery import model
from h2o_discovery._internal import load


def mock_async_client():
    client = mock.Mock()
    client.get_environment.return_value = asyncio.Future()
    client.get_environment.return_value.set_result(mock.Mock())
    client.list_services.return_value = asyncio.Future()
    client.list_services.return_value.set_result({})
    client.list_clients.return_value = asyncio.Future()
    client.list_clients.return_value.set_result({})
    client.list_links.return_value = asyncio.Future()
    client.list_links.return_value.set_result({})
    client.list_components.return_value = asyncio.Future()
    client.list_components.return_value.set_result({})

    return client


def mock_sync_client():
    client = mock.Mock()
    client.get_environment.return_value = {}
    client.list_services.return_value = {}
    client.list_clients.return_value = {}
    client.list_clients.return_value = {}
    client.list_links.return_value = {}
    client.list_components.return_value = {}

    return client


ENVIRONMENT_DATA = model.Environment(
    h2o_cloud_environment="https://test.example.com",
    h2o_cloud_platform_oauth2_scope="test-scope",
    issuer_url="https://test.example.com",
    h2o_cloud_version="test-version",
)


def test_load_environment():
    # Given
    mock_client = mock_sync_client()
    mock_client.get_environment.return_value = ENVIRONMENT_DATA

    # When
    discovery = load.load_discovery(mock_client)

    # Then
    assert discovery.environment == ENVIRONMENT_DATA


@pytest.mark.asyncio
async def test_load_environment_async():
    # Given
    mock_client = mock_async_client()
    mock_client.get_environment.return_value = asyncio.Future()
    mock_client.get_environment.return_value.set_result(ENVIRONMENT_DATA)

    # When
    discovery = await load.load_discovery_async(mock_client)

    # Then
    assert discovery.environment == ENVIRONMENT_DATA


def test_load_services():
    # Given
    mock_client = mock_sync_client()
    mock_client.list_services.return_value = [SERVICE_RECORD]

    # When
    discovery = load.load_discovery(mock_client)

    # Then
    assert discovery.services["test-service"] == SERVICE_RECORD


SERVICE_RECORD = model.Service(
    name="services/test-service",
    display_name="Test Service",
    uri="http://test-service.domain:1234",
    version="1.0.0",
    oauth2_scope="test-service-scope",
    python_client="test-client==1.0.0",
)


@pytest.mark.asyncio
async def test_load_services_async():
    # Given
    mock_client = mock_async_client()
    mock_client.list_services.return_value = asyncio.Future()
    mock_client.list_services.return_value.set_result([SERVICE_RECORD])

    # When
    discovery = await load.load_discovery_async(mock_client)

    # Then
    assert discovery.services["test-service"] == SERVICE_RECORD


CLIENT_RECORD = model.Client(
    name="clients/test-client",
    display_name="Test Client",
    oauth2_client_id="test-client-id",
)


def test_load_clients():
    # Given
    mock_client = mock_sync_client()
    mock_client.list_clients.return_value = [CLIENT_RECORD]

    # When
    discovery = load.load_discovery(mock_client)

    # Then
    assert discovery.clients["test-client"] == CLIENT_RECORD


@pytest.mark.asyncio
async def test_load_clients_async():
    # Given
    mock_client = mock_async_client()
    mock_client.list_clients.return_value = asyncio.Future()
    mock_client.list_clients.return_value.set_result([CLIENT_RECORD])

    # When
    discovery = await load.load_discovery_async(mock_client)

    # Then
    assert discovery.clients["test-client"] == CLIENT_RECORD


LINK_RECORD = model.Link(
    name="links/test-link", uri="http://test-link.domain:1234", text="Test Link"
)


def test_load_links():
    # Given
    mock_client = mock_sync_client()
    mock_client.list_links.return_value = [LINK_RECORD]

    # When
    discovery = load.load_discovery(mock_client)

    # Then
    assert discovery.links["test-link"] == LINK_RECORD


@pytest.mark.asyncio
async def test_load_links_async():
    # Given
    mock_client = mock_async_client()
    mock_client.list_links.return_value = asyncio.Future()
    mock_client.list_links.return_value.set_result([LINK_RECORD])

    # When
    discovery = await load.load_discovery_async(mock_client)

    # Then
    assert discovery.links["test-link"] == LINK_RECORD


def test_load_links_not_found_returns_empty_map():
    # Given
    not_found_exception = httpx.HTTPStatusError(
        "Test Error", request=mock.Mock(), response=mock.Mock(status_code=404)
    )
    mock_client = mock_sync_client()
    mock_client.list_links.side_effect = not_found_exception

    # When
    discovery = load.load_discovery(mock_client)

    # Then
    assert discovery.links == {}


@pytest.mark.asyncio
async def test_load_links_async_not_found_returns_empty_map():
    # Given
    not_found_exception = httpx.HTTPStatusError(
        "Test Error", request=mock.Mock(), response=mock.Mock(status_code=404)
    )
    mock_client = mock_async_client()
    mock_client.list_links.side_effect = not_found_exception

    # When
    discovery = await load.load_discovery_async(mock_client)

    # Then
    assert discovery.links == {}


COMPONENT_RECORD = model.Component(
    name="components/test-component",
    display_name="Test Component",
    description="Test Description",
    version="1.0.0",
)


def test_load_components():
    # Given
    mock_client = mock_sync_client()
    mock_client.list_components.return_value = [COMPONENT_RECORD]

    # When
    discovery = load.load_discovery(mock_client)

    # Then
    assert discovery.components["test-component"] == COMPONENT_RECORD


@pytest.mark.asyncio
async def test_load_components_async():
    # Given
    mock_client = mock_async_client()
    mock_client.list_components.return_value = asyncio.Future()
    mock_client.list_components.return_value.set_result([COMPONENT_RECORD])

    # When
    discovery = await load.load_discovery_async(mock_client)

    # Then
    assert discovery.components["test-component"] == COMPONENT_RECORD


def test_load_components_not_found_returns_empty_map():
    # Given
    not_found_exception = httpx.HTTPStatusError(
        "Test Error", request=mock.Mock(), response=mock.Mock(status_code=404)
    )
    mock_client = mock_sync_client()
    mock_client.list_components.side_effect = not_found_exception

    # When
    discovery = load.load_discovery(mock_client)

    # Then
    assert discovery.components == {}


@pytest.mark.asyncio
async def test_load_components_async_not_found_returns_empty_map():
    # Given
    not_found_exception = httpx.HTTPStatusError(
        "Test Error", request=mock.Mock(), response=mock.Mock(status_code=404)
    )
    mock_client = mock_async_client()
    mock_client.list_components.side_effect = not_found_exception

    # When
    discovery = await load.load_discovery_async(mock_client)

    # Then
    assert discovery.components == {}
